import logging
from typing import Optional

import requests
from soda.__version__ import SODA_CORE_VERSION
from soda.soda_cloud.historic_descriptor import (
    HistoricCheckResultsDescriptor,
    HistoricDescriptor,
    HistoricMeasurementsDescriptor,
)

logger = logging.getLogger(__name__)


class SodaCloudClient:
    def __init__(
        self,
        api_key_id: str,
        api_key_secret: str,
        host: Optional[str] = None,
        port: Optional[int] = None,
        token: Optional[str] = None,
    ):
        self.host = host
        self.port = f":{port}" if port else ""
        self.api_url = f"https://{self.host}{self.port}/api"
        self.api_key_id = api_key_id
        self.api_key_secret = api_key_secret
        self.token: Optional[str] = token
        self.headers = {"User-Agent": f"SodaCore/{SODA_CORE_VERSION}"}

    def insert_scan_results(self, scan_results):
        scan_results["type"] = "sodaCoreInsertScanResults"
        return self._execute_command(scan_results)

    def get_historic_data(self, historic_descriptor: HistoricDescriptor):
        measurements = {}
        check_results = {}

        if type(historic_descriptor) == HistoricMeasurementsDescriptor:
            measurements = self._get_historic_measurements(historic_descriptor)

        elif type(historic_descriptor) == HistoricCheckResultsDescriptor:
            check_results = self._get_hisotric_check_results(historic_descriptor)
        else:
            logger.error(f"Invalid Historic Descriptor provided {historic_descriptor}")

        return {"measurements": measurements, "check_results": check_results}

    def _get_historic_measurements(self, hd: HistoricMeasurementsDescriptor):
        return self._execute_query(
            {
                "type": "sodaCoreHistoricMeasurements",
                "limit": hd.limit,
                "filter": {
                    "type": "and",
                    "andExpressions": [
                        {
                            "type": "equals",
                            "left": {"type": "columnValue", "columnName": "metric.identity"},
                            "right": {"type": "string", "value": hd.metric_identity},
                        }
                    ],
                },
            }
        )

    def _get_hisotric_check_results(self, hd: HistoricCheckResultsDescriptor):
        return self._execute_query(
            {
                "type": "sodaCoreHistoricCheckResults",
                "limit": hd.limit,
                "filter": {
                    "type": "and",
                    "andExpressions": [
                        {
                            "type": "equals",
                            "left": {"type": "columnValue", "columnName": "check.identity"},
                            "right": {"type": "string", "value": hd.check_identity},
                        }
                    ],
                },
            }
        )

    def _execute_query(self, command: dict):
        return self._execute_request("query", command, False)

    def _execute_command(self, command: dict):
        return self._execute_request("command", command, False)

    def _execute_request(self, request_type: str, request_body: dict, is_retry: bool):
        request_body["token"] = self._get_token()
        response = requests.post(f"{self.api_url}/{request_type}", json=request_body, headers=self.headers)
        response_json = response.json()
        if response.status_code == 401 and not is_retry:
            logger.debug(f"Authentication failed. Probably token expired. Re-authenticating...")
            self.token = None
            response_json = self._execute_request(request_type, request_body, True)
        elif response.status_code != 200:
            logger.debug(f"Error while executing Soda cloud request {response_json}")
        return response_json

    def _get_token(self):
        if not self.token:
            login_command = {"type": "login"}
            if self.api_key_id and self.api_key_secret:
                logger.debug("> /api/command (login with API key credentials)")
                login_command["apiKeyId"] = self.api_key_id
                login_command["apiKeySecret"] = self.api_key_secret
            else:
                raise RuntimeError("No API KEY and/or SECRET provided ")

            login_response = requests.post(f"{self.api_url}/command", json=login_command, headers=self.headers)

            if login_response.status_code != 200:
                raise AssertionError(f"< {login_response.status_code} Login failed: {login_response.content}")
            login_response_json = login_response.json()
            self.token = login_response_json.get("token")
            assert self.token, "No token in login response?!"
            logger.debug("< 200 (login ok, token received)")
        return self.token



from __future__ import annotations

import json
import logging
import os
import textwrap
from datetime import datetime

from soda.__version__ import SODA_CORE_VERSION
from soda.common.log import Log, LogLevel
from soda.common.logs import Logs
from soda.common.undefined_instance import undefined
from soda.execution.check import Check
from soda.execution.check_outcome import CheckOutcome
from soda.execution.data_source_scan import DataSourceScan
from soda.execution.derived_metric import DerivedMetric
from soda.execution.metric import Metric
from soda.soda_cloud.historic_descriptor import HistoricDescriptor
from soda.sodacl.location import Location
from soda.sodacl.sodacl_cfg import SodaCLCfg

logger = logging.getLogger(__name__)
verbose = False


class Scan:
    def __init__(self):
        from soda.configuration.configuration import Configuration
        from soda.execution.check import Check
        from soda.execution.data_source_manager import DataSourceManager
        from soda.execution.query import Query

        now = datetime.utcnow()
        self._logs = Logs(logger)
        self._scan_definition_name: str | None = None
        self._data_source_name: str | None = None
        self._variables: dict[str, object] = {"NOW": now.isoformat()}
        self._configuration: Configuration = Configuration(scan=self)
        self._sodacl_cfg: SodaCLCfg = SodaCLCfg(scan=self)
        self._file_paths: set[str] = set()
        self._data_timestamp: datetime = now
        self._scan_start_timestamp: datetime = now
        # FIXME: this attribute cannot be None if typed as `datetime`
        self._scan_end_timestamp: datetime | None = None
        self._data_source_manager = DataSourceManager(self._logs, self._configuration)
        self._data_source_scans: list[DataSourceScan] = []
        self._metrics: set[Metric] = set()
        self._checks: list[Check] = []
        self._queries: list[Query] = []
        self._logs.info(f"Soda Core {SODA_CORE_VERSION}")

    def set_data_source_name(self, data_source_name: str):
        """
        Specifies which datasource to use for the checks.
        """
        self._data_source_name = data_source_name

    def set_scan_definition_name(self, scan_definition_name: str):
        """
        The scan definition name is required if the scan is connected to Soda Cloud in order to correlate subsequent scans from the same pipeline.
        """
        self._scan_definition_name = scan_definition_name

    def set_verbose(self, verbose_var: bool = True):
        self._logs.verbose = verbose_var
        global verbose
        verbose = verbose_var

    def add_configuration_yaml_file(self, file_path: str):
        """
        Configure environment configurations from a YAML file
        """
        try:
            configuration_yaml_str = self._read_file("configuration", file_path)
            self._parse_configuration_yaml_str(
                configuration_yaml_str=configuration_yaml_str,
                file_path=file_path,
            )
        except Exception as e:
            self._logs.error(
                f"Could not add environment configurations from file path {file_path}",
                exception=e,
            )

    def add_configuration_yaml_str(self, environment_yaml_str: str, file_path: str = "yaml string"):
        """
        Configure environment configurations from the given string.
        Parameter file_path is optional and can be used to get the location of the log/error in the logs.
        """
        try:
            self._parse_configuration_yaml_str(
                configuration_yaml_str=environment_yaml_str,
                file_path=file_path,
            )
        except Exception as e:
            self._logs.error(
                f"Could not add environment configurations from string",
                exception=e,
            )

    def _parse_configuration_yaml_str(self, configuration_yaml_str: str, file_path: str = "yaml string"):
        from soda.configuration.configuration_parser import ConfigurationParser

        environment_parse = ConfigurationParser(
            configuration=self._configuration,
            logs=self._logs,
            file_path=file_path,
        )
        environment_parse.parse_environment_yaml_str(configuration_yaml_str)

    def add_configuration_spark_session(self, data_source_name: str, spark_session):
        """
        Pass a spark_session to the scan.  Only required in case of PySpark scans.
        """
        try:
            self._configuration.add_spark_session(data_source_name=data_source_name, spark_session=spark_session)
        except Exception as e:
            self._logs.error(
                f"Could not add environment spark session for data_source {data_source_name}",
                exception=e,
            )

    def add_sodacl_yaml_files(
        self,
        path: str,
        recursive: bool | None = True,
        suffix: str | None = ".yml",
    ):
        """
        Adds all the files in the given directory to the scan as SodaCL files.
        Parameter 'path' is a string that typically represents a directory, but it can also be a SodaCL file.
                         ~ will be expanded to the user home dir the directory in which to search for SodaCL files.
        Parameter 'recursive' controls if nested directories also will be scanned.  Default recursive=True.
        Parameter 'suffix' can be used to only load files having a given extension or suffix like eg suffix='.sodacl.yml'
        """
        try:
            if isinstance(path, str):
                file_system = self._configuration.file_system
                path = file_system.expand_user(path)
                if file_system.exists(path):
                    if file_system.is_dir(path):
                        self._logs.info(f"Adding SodaCL dir {path}")
                        for dir_entry in file_system.scan_dir(path):
                            if dir_entry.is_file() and (suffix is None or dir_entry.name.endswith(suffix)):
                                self.add_sodacl_yaml_file(file_path=dir_entry.path)
                            elif recursive and dir_entry.is_dir():
                                self.add_sodacl_yaml_files(
                                    path=dir_entry.path,
                                    recursive=True,
                                    suffix=suffix,
                                )
                    elif file_system.is_file(path):
                        self.add_sodacl_yaml_file(file_path=path)
                    else:
                        self._logs.error(f'path "{path}" exists, but is not a file nor directory ?!')
                else:
                    self._logs.error(f'path "{path}" does not exist')
            else:
                self._logs.error(f"path is not a string: {type(path).__name__}")
        except Exception as e:
            self._logs.error(f"Could not add SodaCL files from dir {dir}", exception=e)

    def add_sodacl_yaml_file(self, file_path: str):
        """
        Add a SodaCL file to the scan.
        """
        try:
            sodacl_yaml_str = self._read_file("SodaCL", file_path)
            if file_path not in self._file_paths:
                self._file_paths.add(file_path)
                self._parse_sodacl_yaml_str(sodacl_yaml_str=sodacl_yaml_str, file_path=file_path)
            else:
                self._logs.debug(f"Skipping duplicate file addition for {file_path}")
        except Exception as e:
            self._logs.error(f"Could not add SodaCL file {file_path}", exception=e)

    def add_sodacl_yaml_str(self, sodacl_yaml_str: str):
        """
        Add a SodaCL string to the scan.
        """
        try:
            unique_name = "sodacl_string"
            if unique_name in self._file_paths:
                number: int = 2
                while f"{unique_name}_{number}" in self._file_paths:
                    number += 1
                unique_name = f"{unique_name}_{number}"
            file_path = f"{unique_name}.yml"
            self._parse_sodacl_yaml_str(sodacl_yaml_str=sodacl_yaml_str, file_path=file_path)
        except Exception as e:
            self._logs.error(f"Could not add SodaCL string", exception=e)

    def _parse_sodacl_yaml_str(self, sodacl_yaml_str: str, file_path: str = None):
        from soda.sodacl.sodacl_parser import SodaCLParser

        sodacl_parser = SodaCLParser(
            sodacl_cfg=self._sodacl_cfg,
            logs=self._logs,
            file_path=file_path,
            data_source_name=self._data_source_name,
        )
        sodacl_parser.parse_sodacl_yaml_str(sodacl_yaml_str)

    def _read_file(self, file_type: str, file_path: str) -> str:
        file_location = Location(file_path)
        file_system = self._configuration.file_system
        resolved_file_path = file_system.expand_user(file_path)
        if not file_system.exists(resolved_file_path):
            self._logs.error(
                f"File {resolved_file_path} does not exist",
                location=file_location,
            )
            return None
        if file_system.is_dir(resolved_file_path):
            self._logs.error(
                f"File {resolved_file_path} exists, but is a directory",
                location=file_location,
            )
            return None
        try:
            self._logs.debug(f'Reading {file_type} file "{resolved_file_path}"')
            file_content_str = file_system.file_read_as_str(resolved_file_path)
            if not isinstance(file_content_str, str):
                self._logs.error(
                    f"Error reading file {resolved_file_path} from the file system",
                    location=file_location,
                )
            return file_content_str
        except Exception as e:
            self._logs.error(
                f"Error reading file {resolved_file_path} from the file system",
                location=file_location,
                exception=e,
            )

    def add_variables(self, variables: dict[str, str]):
        """
        Add variables to the scan. Keys and values must be strings.
        """
        try:
            self._variables.update(variables)
        except Exception as e:
            variables_text = json.dumps(variables)
            self._logs.error(f"Could not add variables {variables_text}", exception=e)

    def disable_telemetry(self):
        """
        Disables all telemetry.  For more information see Soda's public statements on telemetry.  TODO add links.
        """
        self._configuration.telemetry = None

    def execute(self):
        self._logs.debug("Scan execution starts")
        try:
            from soda.execution.column import Column
            from soda.execution.column_metrics import ColumnMetrics
            from soda.execution.partition import Partition
            from soda.execution.table import Table

            # If there is a sampler
            if self._configuration.sampler:
                # ensure the sampler is configured with the scan logs
                self._configuration.sampler.logs = self._logs

            # Resolve the for each table checks and add them to the scan_cfg data structures
            self.__resolve_for_each_table_checks()
            # Resolve the for each column checks and add them to the scan_cfg data structures
            self.__resolve_for_each_column_checks()

            # For each data_source, build up the DataSourceScan data structures
            for data_source_scan_cfg in self._sodacl_cfg.data_source_scan_cfgs.values():
                # This builds up the data structures that correspond to the cfg model
                data_source_scan = self._get_or_create_data_source_scan(data_source_scan_cfg.data_source_name)
                if data_source_scan:
                    for check_cfg in data_source_scan_cfg.check_cfgs:
                        self.__create_check(check_cfg, data_source_scan)

                    for table_cfg in data_source_scan_cfg.tables_cfgs.values():
                        table: Table = data_source_scan.get_or_create_table(table_cfg.table_name)

                        for column_configurations_cfg in table_cfg.column_configurations_cfgs.values():
                            column: Column = table.get_or_create_column(column_configurations_cfg.column_name)
                            column.set_column_configuration_cfg(column_configurations_cfg)

                        for partition_cfg in table_cfg.partition_cfgs:
                            partition: Partition = table.get_or_create_partition(partition_cfg.partition_name)
                            partition.set_partition_cfg(partition_cfg)

                            for check_cfg in partition_cfg.check_cfgs:
                                self.__create_check(check_cfg, data_source_scan, partition)

                            if partition_cfg.column_checks_cfgs:
                                for column_checks_cfg in partition_cfg.column_checks_cfgs.values():
                                    column_metrics: ColumnMetrics = partition.get_or_create_column_metrics(
                                        column_checks_cfg.column_name
                                    )
                                    column_metrics.set_column_check_cfg(column_checks_cfg)
                                    if column_checks_cfg.check_cfgs:
                                        for check_cfg in column_checks_cfg.check_cfgs:
                                            self.__create_check(
                                                check_cfg,
                                                data_source_scan,
                                                partition,
                                                column_metrics.column,
                                            )

            # Each data_source is asked to create metric values that are returned as a list of query results
            for data_source_scan in self._data_source_scans:
                data_source_scan.execute_queries()

            # Compute derived metric values
            for metric in self._metrics:
                if isinstance(metric, DerivedMetric):
                    metric.compute_derived_metric_values()

            # Evaluates the checks based on all the metric values
            for check in self._checks:
                # First get the metric values for this check
                check_metrics = {}
                missing_value_metrics = []
                for check_metric_name, metric in check.metrics.items():
                    if metric.value is not undefined:
                        check_metrics[check_metric_name] = metric
                    else:
                        missing_value_metrics.append(metric)

                check_historic_data = {}
                # For each check get the historic data
                if check.historic_descriptors:
                    for hd_key, hd in check.historic_descriptors.items():
                        check_historic_data[hd_key] = self.__get_historic_data_from_soda_cloud_metric_store(hd)

                if not missing_value_metrics:
                    try:
                        check.evaluate(check_metrics, check_historic_data)
                    except BaseException as e:
                        self._logs.error(
                            f"Evaluation of check {check.check_cfg.source_line} failed: {e}",
                            location=check.check_cfg.location,
                            exception=e,
                        )
                else:
                    missing_metrics_str = ",".join([str(metric) for metric in missing_value_metrics])
                    self._logs.error(
                        f"Metrics {missing_metrics_str} were not computed for check {check.check_cfg.source_line}"
                    )

            for data_source_scan in self._data_source_scans:
                for monitoring_cfg in data_source_scan.data_source_scan_cfg.monitoring_cfgs:
                    data_source_name = data_source_scan.data_source_scan_cfg.data_source_name
                    data_source_scan = self._get_or_create_data_source_scan(data_source_name)
                    if data_source_scan:
                        monitor_runner = data_source_scan.create_automated_monitor_run(monitoring_cfg, self)
                        monitor_runner.run()
                    else:
                        data_source_names = ", ".join(self._data_source_manager.data_source_properties_by_name.keys())
                        self._logs.error(
                            f"Could not run monitors on data_source {data_source_name} because It is not "
                            f"configured: {data_source_names}"
                        )

            self._logs.info("Scan summary:")
            self.__log_queries(having_exception=False)
            self.__log_queries(having_exception=True)

            checks_pass_count = self.__log_checks(CheckOutcome.PASS)
            checks_warn_count = self.__log_checks(CheckOutcome.WARN)
            warn_text = "warning" if checks_warn_count == 1 else "warnings"
            checks_fail_count = self.__log_checks(CheckOutcome.FAIL)
            fail_text = "failure" if checks_warn_count == 1 else "failures"
            error_count = len(self.get_error_logs())
            error_text = "error" if error_count == 1 else "errors"
            self.__log_checks(None)
            checks_not_evaluated = len(self._checks) - checks_pass_count - checks_warn_count - checks_fail_count

            if checks_not_evaluated:
                self._logs.info(f"{checks_not_evaluated} checks not evaluated.")
            if error_count > 0:
                self._logs.info(f"{error_count} errors.")
            if checks_warn_count + checks_fail_count + error_count == 0:
                if checks_not_evaluated:
                    self._logs.info(
                        f"Apart from the checks that have not been evaluated, no failures, no warnings and no errors."
                    )
                else:
                    self._logs.info(f"All is good. No failures. No warnings. No errors.")
            elif checks_fail_count > 0:
                self._logs.info(
                    f"Oops! {checks_fail_count} {fail_text}. {checks_warn_count} {warn_text}. {error_count} {error_text}. {checks_pass_count} pass."
                )
            elif checks_warn_count > 0:
                self._logs.info(
                    f"Only {checks_warn_count} {warn_text}. {checks_fail_count} {fail_text}. {error_count} {error_text}. {checks_pass_count} pass."
                )
            elif error_count > 0:
                self._logs.info(
                    f"Oops! {error_count} {error_text}. {checks_fail_count} {fail_text}. {checks_warn_count} {warn_text}. {checks_pass_count} pass."
                )

            if error_count > 0:
                Log.log_errors(self.get_error_logs())

            self._scan_end_timestamp = datetime.utcnow()
            if self._configuration.soda_cloud:
                self._logs.info("Sending results to Soda Cloud")
                self._configuration.soda_cloud.send_scan_results(self)

        except Exception as e:
            self._logs.error(f"Error occurred while executing scan.", exception=e)
        finally:
            self._close()

    def __checks_to_text(self, checks: list[Check]):
        return "/n".join([str(check) for check in checks])

    def _close(self):
        self._data_source_manager.close_all_connections()

    def __create_check(self, check_cfg, data_source_scan=None, partition=None, column=None):
        from soda.execution.check import Check

        check = Check.create(
            check_cfg=check_cfg,
            data_source_scan=data_source_scan,
            partition=partition,
            column=column,
        )
        self._checks.append(check)

    def __resolve_for_each_table_checks(self):
        from soda.execution.query import Query

        data_source_name = self._data_source_name

        for index, for_each_table_cfg in enumerate(self._sodacl_cfg.for_each_table_cfgs):
            include_tables = [include.table_name_filter for include in for_each_table_cfg.includes]
            exclude_tables = [include.table_name_filter for include in for_each_table_cfg.excludes]

            data_source_scan = self._get_or_create_data_source_scan(data_source_name)
            if data_source_scan:
                query_name = f"for_each_table_{for_each_table_cfg.table_alias_name}[{index}]"
                # TODO: use get of data source instead of executing it here (once its available)
                sql = data_source_scan.data_source.sql_find_table_names(
                    include_tables=include_tables, exclude_tables=exclude_tables
                )
                query = Query(data_source_scan=data_source_scan, unqualified_query_name=query_name, sql=sql)
                query.execute()
                table_names = [row[0] for row in query.rows]

                for table_name in table_names:
                    data_source_scan_cfg = self._sodacl_cfg._get_or_create_data_source_scan_cfgs(data_source_name)
                    table_cfg = data_source_scan_cfg.get_or_create_table_cfg(table_name)
                    partition_cfg = table_cfg.find_partition(None, None)
                    for check_cfg in for_each_table_cfg.check_cfgs:
                        column_name = check_cfg.get_column_name()
                        if column_name:
                            column_checks_cfg = partition_cfg.get_or_create_column_checks(column_name)
                            column_checks_cfg.add_check_cfg(check_cfg)
                        else:
                            partition_cfg.add_check_cfg(check_cfg)

    def __resolve_for_each_column_checks(self):
        if self._sodacl_cfg.for_each_column_cfgs:
            raise NotImplementedError("TODO")

    def _get_or_create_data_source_scan(self, data_source_name: str) -> DataSourceScan:
        from soda.execution.data_source import DataSource
        from soda.sodacl.data_source_scan_cfg import DataSourceScanCfg

        data_source_scan = next(
            (
                data_source_scan
                for data_source_scan in self._data_source_scans
                if data_source_scan.data_source.data_source_name == data_source_name
            ),
            None,
        )
        if data_source_scan is None:
            data_source_scan_cfg = self._sodacl_cfg.data_source_scan_cfgs.get(data_source_name)
            if data_source_scan_cfg is None:
                data_source_scan_cfg = DataSourceScanCfg(data_source_name)
            data_source_name = data_source_scan_cfg.data_source_name
            data_source: DataSource = self._data_source_manager.get_data_source(data_source_name)
            if data_source:
                data_source_scan = data_source.create_data_source_scan(self, data_source_scan_cfg)
                self._data_source_scans.append(data_source_scan)
            else:
                self._sodacl_cfg.data_source_scan_cfgs.pop(data_source_name)
        return data_source_scan

    def _jinja_resolve(
        self,
        definition: str,
        variables: dict[str, object] = None,
        location: Location | None = None,
    ):
        if isinstance(definition, str) and "${" in definition:
            from soda.common.jinja import Jinja

            jinja_variables = self._variables.copy()
            if isinstance(variables, dict):
                jinja_variables.update(variables)
            try:
                return Jinja.resolve(definition, jinja_variables)
            except BaseException as e:
                self._logs.error(
                    message=f"Error resolving Jinja template {definition}: {e}",
                    location=location,
                    exception=e,
                )
        else:
            return definition

    def __get_historic_data_from_soda_cloud_metric_store(
        self, historic_descriptor: HistoricDescriptor
    ) -> dict[str, object]:
        if self._configuration.soda_cloud:
            return self._configuration.soda_cloud.get_historic_data(historic_descriptor)
        else:
            self._logs.error("Soda Core must be configured to connect to Soda Cloud to use change-over-time checks.")
        return {}

    def _find_existing_metric(self, metric) -> Metric:
        return next(
            (existing_metric for existing_metric in self._metrics if existing_metric == metric),
            None,
        )

    def _add_metric(self, metric):
        self._metrics.add(metric)

    def __log_queries(self, having_exception: bool) -> int:
        count = sum((query.exception is None) != having_exception for query in self._queries)
        if count > 0:
            status_text = "ERROR" if having_exception else "OK"
            queries_text = "query" if len(self._queries) == 1 else "queries"
            self._logs.debug(f"{count}/{len(self._queries)} {queries_text} {status_text}")
            for query in self._queries:
                query_text = f"\n{query.sql}" if query.exception else ""
                self._logs.debug(f"  {query.query_name} [{status_text}] {query.duration}{query_text}")
                if query.exception:
                    exception_str = str(query.exception)
                    exception_str = textwrap.indent(text=exception_str, prefix="    ")
                    self._logs.debug(exception_str)
        return count

    def __log_checks(self, check_outcome: CheckOutcome | None) -> int:
        count = sum(check.outcome == check_outcome for check in self._checks)
        if count > 0:
            outcome_text = "NOT EVALUATED" if check_outcome is None else f"{check_outcome.value.upper()}ED"
            checks_text = "check" if len(self._checks) == 1 else "checks"
            self._logs.info(f"{count}/{len(self._checks)} {checks_text} {outcome_text}: ")

            checks_by_partition = {}
            other_checks = []
            for check in self._checks:
                if check.outcome == check_outcome:
                    partition = check.partition
                    if partition:
                        partition_name = f" [{partition.partition_name}]" if partition.partition_name else ""
                        partition_title = f"{partition.table.table_name}{partition_name} in {partition.data_source_scan.data_source.data_source_name}"
                        checks_by_partition.setdefault(partition_title, []).append(check)
                    else:
                        other_checks.append(check)

            for (
                partition_title,
                partition_checks,
            ) in checks_by_partition.items():
                if len(partition_checks) > 0:
                    self._logs.info(f"    {partition_title}")
                    self.__log_check_group(partition_checks, "      ", check_outcome, outcome_text)
            if len(other_checks) > 0:
                self.__log_check_group(other_checks, "    ", check_outcome, outcome_text)
        return count

    def __log_check_group(self, checks, indent, check_outcome, outcome_text):
        for check in checks:
            self._logs.info(f"{indent}{check.get_summary()} [{outcome_text}]")
            if self._logs.verbose or check_outcome != CheckOutcome.PASS:
                for diagnostic in check.get_log_diagnostic_lines():
                    self._logs.info(f"{indent}  {diagnostic}")

    def get_variable(self, variable_name: str, default_value: str | None = None) -> str | None:
        # Note: ordering here must be the same as in Jinja.OsContext.resolve_or_missing: First env vars, then scan vars
        if variable_name in os.environ:
            return os.environ[variable_name]
        elif variable_name in self._variables:
            return self._variables[variable_name]
        return default_value

    def get_logs_text(self) -> str | None:
        return self.__logs_to_text(self._logs.logs)

    def has_error_logs(self) -> bool:
        return any(log.level == LogLevel.ERROR for log in self._logs.logs)

    def get_error_logs(self) -> list[Log]:
        return [log for log in self._logs.logs if log.level == LogLevel.ERROR]

    def get_error_logs_text(self) -> str | None:
        return self.__logs_to_text(self.get_error_logs())

    def assert_no_error_logs(self) -> None:
        if self.has_error_logs():
            raise AssertionError(self.get_error_logs_text())

    def has_error_or_warning_logs(self) -> bool:
        return any(log.level in [LogLevel.ERROR, LogLevel.WARNING] for log in self._logs.logs)

    def get_error_or_warning_logs(self) -> list[Log]:
        return [log for log in self._logs.logs if log.level in [LogLevel.ERROR, LogLevel.WARNING]]

    def get_error_or_warning_logs_text(self) -> str | None:
        return self.__logs_to_text(self.get_error_or_warning_logs())

    def assert_no_error_nor_warning_logs(self) -> None:
        if self.has_error_or_warning_logs():
            raise AssertionError(self.get_logs_text())

    def assert_has_error(self, expected_error_message: str):
        if all(
            [
                expected_error_message not in log.message and expected_error_message not in str(log.exception)
                for log in self.get_error_logs()
            ]
        ):
            raise AssertionError(
                f'Expected error message "{expected_error_message}" did not occur in the error logs:\n{self.get_logs_text()}'
            )

    def __logs_to_text(self, logs: list[Log]):
        if len(logs) == 0:
            return None
        return "\n".join([str(log) for log in logs])

    def has_check_fails(self) -> bool:
        for check in self._checks:
            if check.outcome == CheckOutcome.FAIL:
                return True
        return False

    def has_check_warns(self) -> bool:
        for check in self._checks:
            if check.outcome == CheckOutcome.WARN:
                return True
        return False

    def has_check_warns_or_fails(self) -> bool:
        for check in self._checks:
            if check.outcome in [CheckOutcome.FAIL, CheckOutcome.WARN]:
                return True
        return False

    def assert_no_checks_fail(self):
        if len(self.get_checks_fail()):
            raise AssertionError(f"Check results failed: \n{self.get_checks_fail_text()}")

    def get_checks_fail(self) -> list[Check]:
        return [check for check in self._checks if check.outcome == CheckOutcome.FAIL]

    def get_checks_fail_text(self) -> str | None:
        return self.__checks_to_text(self.get_checks_fail())

    def assert_no_checks_warn_or_fail(self):
        if len(self.get_checks_warn_or_fail()):
            raise AssertionError(f"Check results having warn or fail outcome: \n{self.get_checks_warn_or_fail_text()}")

    def get_checks_warn_or_fail(self) -> list[Check]:
        return [check for check in self._checks if check.outcome in [CheckOutcome.WARN, CheckOutcome.FAIL]]

    def has_checks_warn_or_fail(self) -> bool:
        return len(self.get_checks_warn_or_fail()) > 0

    def get_checks_warn_or_fail_text(self) -> str | None:
        return self.__checks_to_text(self.get_checks_warn_or_fail())

    def get_all_checks_text(self) -> str | None:
        return self.__checks_to_text(self._checks)

    def has_soda_cloud_connection(self):
        return self._configuration.soda_cloud is not None



from typing import Callable, Dict


class DerivedFormula:
    def __init__(self, function: Callable, metric_dependencies: Dict[str, "Metric"]):
        self.function: Callable = function
        self.metric_dependencies: Dict[str, "Metric"] = metric_dependencies



from __future__ import annotations

from soda.execution.aggregation_query import AggregationQuery
from soda.execution.duplicates_query import DuplicatesQuery
from soda.execution.numeric_query_metric import NumericQueryMetric
from soda.execution.query import Query
from soda.execution.schema_query import SchemaQuery
from soda.sodacl.partition_cfg import PartitionCfg


class Partition:
    def __init__(self, table: Table, partition_name: str):
        from soda.execution.column_metrics import ColumnMetrics
        from soda.execution.metric import Metric
        from soda.execution.table import Table
        from soda.sodacl.partition_cfg import PartitionCfg

        self.logs = table.data_source_scan.scan._logs
        self.table: Table = table
        self.partition_name = partition_name
        self.partition_cfg: PartitionCfg = None
        self.sql_partition_filter: str | None = None
        self.data_source_scan = table.data_source_scan
        self.metrics: list[Metric] = []
        self.column_metrics: dict[str, ColumnMetrics] = {}

        self.schema_query: SchemaQuery | None = None
        self.aggregation_queries: list[AggregationQuery] = []
        self.duplicate_queries: list[DuplicatesQuery] = []

    def set_partition_cfg(self, partition_cfg: PartitionCfg):
        self.partition_cfg = partition_cfg
        self.sql_partition_filter = partition_cfg.sql_partition_filter

    def add_check_cfg(self, check_cfg: CheckCfg):
        if self.partition_cfg is None:
            self.partition_cfg = PartitionCfg(self.partition_name)
        self.partition_cfg.add_check_cfg(check_cfg)

    def get_or_create_column_metrics(self, column_name: str):
        from soda.execution.column_metrics import ColumnMetrics

        column_metrics = self.column_metrics.get(column_name)
        if not column_metrics:
            column_metrics = ColumnMetrics(self, column_name)
            self.column_metrics[column_name] = column_metrics
        return column_metrics

    def ensure_query_for_metric(self, metric: Metric):
        if isinstance(metric, NumericQueryMetric):
            if metric.name == "duplicate_count":
                duplicates_query = DuplicatesQuery(self, metric)
                self.duplicate_queries.append(duplicates_query)
                metric.queries.append(duplicates_query)
            else:
                sql_aggregation_expression = metric.get_sql_aggregation_expression()
                if sql_aggregation_expression:
                    max_aggregation_fields = self.data_source_scan.data_source.get_max_aggregation_fields()
                    if (
                        len(self.aggregation_queries) == 0
                        or len(self.aggregation_queries[-1].metrics) >= max_aggregation_fields
                    ):
                        aggregation_query_index = len(self.aggregation_queries)
                        aggregation_query = AggregationQuery(self, aggregation_query_index)
                        self.aggregation_queries.append(aggregation_query)
                    else:
                        aggregation_query = self.aggregation_queries[-1]
                    aggregation_query.add_metric(sql_aggregation_expression, metric)
                    metric.queries.append(aggregation_query)
                else:
                    self.logs.error(f"Unsupported metric {metric.name}")
        else:
            self.logs.error(f"Unsupported metric {metric.name} ({type(metric).__name__})")

    def collect_queries(self) -> list[Query]:
        queries: list[Query] = []
        if self.schema_query:
            queries.append(self.schema_query)
        queries.extend(self.aggregation_queries)
        queries.extend(self.duplicate_queries)
        return queries

    @classmethod
    def get_partition_name(cls, partition):
        return partition.partition_name if isinstance(partition, Partition) else None

    @classmethod
    def get_table_name(cls, partition):
        return partition.table.table_name if isinstance(partition, Partition) else None



from soda.execution.query import Query


class UserDefinedFailedRowsQuery(Query):
    def __init__(
        self,
        data_source_scan: "DataSourceScan",
        metric: "FailedRowsQueryMetric",
    ):
        super().__init__(
            data_source_scan=data_source_scan,
            unqualified_query_name=f"failed_rows[{metric.name}]",
        )
        self.sql = metric.query
        self.metric = metric

    def execute(self):
        self.store()
        if self.storage_ref:
            self.metric.value = self.storage_ref.total_row_count
            self.metric.failed_rows_storage_ref = self.storage_ref



from typing import List

from soda.sodacl.check_cfg import CheckCfg
from soda.sodacl.location import Location
from soda.sodacl.name_filter import NameFilter


class ForEachTableCfg:
    def __init__(self):
        self.table_alias_name: str = None
        self.includes: List[NameFilter] = []
        self.excludes: List[NameFilter] = []
        self.check_cfgs: List[CheckCfg] = []
        self.location: Location = None



from abc import ABC
from typing import Optional

from soda.execution.query import Query
from soda.sampler.storage_ref import StorageRef


class Sampler(ABC):
    def __init__(self):
        # Initialized in the scan.execute
        self.logs = None

    def store_sample(self, cursor, query: Optional[Query]) -> Optional[StorageRef]:
        pass



import os
from typing import Optional

from jinja2 import Environment
from jinja2.runtime import Context


class OsContext(Context):
    def resolve_or_missing(self, key):
        # Note: ordering here must be the same as in scan.get_variable(variable_name): First env vars, then scan vars
        if key in os.environ:
            return os.environ[key]
        elif key in self.vars:
            return self.vars[key]
        elif key == "env_var":
            return Jinja.env_var
        return Context.resolve_or_missing(self, key)


def create_os_environment():
    environment = Environment(variable_start_string="${", variable_end_string="}")
    environment.context_class = OsContext
    return environment


class Jinja:

    environment = create_os_environment()

    @staticmethod
    def resolve(template: str, variables: dict = None) -> str:
        """
        Convenience method that funnels Jinja exceptions into parselog errors.
        This method throws no exceptions.  Returns None in case of Jinja exceptions.
        """
        if not isinstance(variables, dict):
            variables = {}
        jinja_template = Jinja.environment.from_string(template)
        rendered_value = jinja_template.render(variables)
        return rendered_value

    @staticmethod
    def env_var(variable_name: str, default_value: Optional[str] = None) -> str:
        """
        Function that can be used in the variables in resolve.  Usage:
        variables = {
            'env_var': Jinja.env_var
        }
        Jinja.resolve(template, variables, parse)
        """
        value = os.getenv(variable_name, default_value)
        return value if isinstance(value, str) else ""



from __future__ import annotations

from abc import abstractmethod

from soda.sodacl.location import Location


class CheckCfg:
    def __init__(
        self,
        source_header: str,
        source_line: str,
        source_configurations: dict | None,
        location: Location,
        name: str | None,
    ):
        self.source_header: str = source_header
        self.source_line: str = source_line
        self.source_configurations = source_configurations
        self.location: Location = location
        self.name: str | None = name

    def get_column_name(self) -> str | None:
        pass

    @abstractmethod
    def get_identity_parts(self) -> list:
        pass



from typing import List

from soda.sodacl.check_cfg import CheckCfg


class ColumnChecksCfg:
    def __init__(self, column_name: str):
        self.column_name = column_name
        self.check_cfgs: List[CheckCfg] = []

    def add_check_cfg(self, check_cfg: CheckCfg):
        self.check_cfgs.append(check_cfg)



from typing import Optional

from soda.execution.identity import Identity
from soda.sodacl.check_cfg import CheckCfg
from soda.sodacl.location import Location


class RowCountComparisonCheckCfg(CheckCfg):
    def __init__(
        self,
        source_header: str,
        source_line: str,
        source_configurations: Optional[str],
        location: Location,
        name: Optional[str],
        other_table_name: str,
        other_partition_name: Optional[str],
        other_data_source_name: Optional[str],
    ):
        super().__init__(source_header, source_line, source_configurations, location, name)
        self.other_table_name: str = other_table_name
        self.other_partition_name: Optional[str] = other_partition_name
        self.other_data_source_name: Optional[str] = other_data_source_name

    def get_identity_parts(self) -> list:
        return [
            self.location,
            self.other_table_name,
            Identity.property("other_partition_name", self.other_partition_name),
            Identity.property("other_data_source_name", self.other_data_source_name),
        ]



from typing import Optional

from soda.sodacl.check_cfg import CheckCfg
from soda.sodacl.location import Location


class UserDefinedFailedRowsCheckCfg(CheckCfg):
    def __init__(
        self,
        source_header: str,
        source_line: str,
        source_configurations: Optional[dict],
        location: Location,
        name: Optional[str],
        query: Optional[str],
    ):
        super().__init__(source_header, source_line, source_configurations, location, name)
        self.query: Optional[str] = query

    def get_identity_parts(self) -> list:
        return [self.location, self.query]



from typing import List

from soda.execution.data_source import DataSource


class ProvidedSparkSessionCursor:
    def __init__(self, connection):
        self.spark_session = connection.spark_session
        self.result_spark_dataframe = None
        self.description = None

    def execute(self, sql: str):
        self.result_spark_dataframe = self.spark_session.sql(sql)

    def fetchone(self):
        spark_rows: List["pyspark.sql.types.Row"] = self.result_spark_dataframe.collect()
        spark_row = spark_rows[0]
        values = [spark_row[field] for field in spark_row.__fields__]
        self.description = tuple((field, "?", True) for field in spark_row.__fields__)
        return tuple(values)

    def close(self):
        pass


class ProvidedSparkSessionConnection:
    def __init__(self, spark_session):
        self.spark_session = spark_session

    def cursor(self):
        return ProvidedSparkSessionCursor(self)

    def rollback(self):
        pass


class DataSourceImpl(DataSource):
    def connect(self, connection_properties: dict):
        provided_spark_session = connection_properties["provided_spark_session"]
        self.connection = ProvidedSparkSessionConnection(provided_spark_session)

    def validate_configuration(self, connection_properties: dict, logs: "Logs") -> None:
        pass



import logging
from abc import ABC
from typing import Optional

from soda.execution.metric import Metric

logger = logging.getLogger(__name__)


class QueryMetric(Metric, ABC):
    def __init__(
        self,
        data_source_scan: "DataSourceScan",
        partition: Optional["Partition"],
        column: Optional["Column"],
        name: str,
        check: "Check",
        identity_parts: list = None,
    ):
        super().__init__(
            data_source_scan=data_source_scan,
            partition=partition,
            column=column,
            name=name,
            check=check,
            identity_parts=identity_parts,
        )



from typing import List

from soda.anomaly_detection.anomaly_detector import AnomalyOutput
from soda.execution.schema_comparator import SchemaComparator
from soda.sodacl.automated_monitoring_cfg import AutomatedMonitoringCfg


class AutomatedMonitoringResult:
    def __init__(self, automated_monitoring_cfg: AutomatedMonitoringCfg):
        self.automated_monitoring_cfg: AutomatedMonitoringCfg = automated_monitoring_cfg

    def get_cloud_diagnostics_dict(self):
        return {}

    def get_log_diagnostic_dict(self) -> dict:
        return {}

    def append_row_count_anomaly_evaluation_result(self, table_name: str, anomaly_output: AnomalyOutput):
        pass

    def append_table_schema_changes(self, schema_comparator: SchemaComparator):
        pass

    def append_table_changes(self, tables_added: List[str], tables_removed: List[str]):
        pass



from typing import Dict, List

from soda.anomaly_detection.anomaly_detector import AnomalyInput, AnomalyOutput
from soda.execution.automated_monitoring_result import AutomatedMonitoringResult
from soda.execution.data_source_scan import DataSourceScan
from soda.execution.query import Query
from soda.execution.schema_comparator import SchemaComparator
from soda.sodacl.automated_monitoring_cfg import AutomatedMonitoringCfg


class AutomatedMonitoringRun:
    def __init__(self, data_source_scan: DataSourceScan, automated_monitoring_cfg: AutomatedMonitoringCfg):
        self.data_source_scan: DataSourceScan = data_source_scan
        self.soda_cloud = data_source_scan.scan._configuration.soda_cloud
        self.data_source = data_source_scan.data_source
        self.automated_monitoring_cfg: AutomatedMonitoringCfg = automated_monitoring_cfg
        self.field_tablename = '"tablename"'
        self.logs = self.data_source_scan.scan._logs

    def run(self) -> AutomatedMonitoringResult:
        automated_monitoring_result: AutomatedMonitoringResult = AutomatedMonitoringResult(
            self.automated_monitoring_cfg
        )

        if self.automated_monitoring_cfg.row_count:
            # row_counts is a dict that maps table names to row counts.
            row_counts_by_table_name: Dict[str, int] = self.get_row_counts_all_tables()
            for measured_table_name in row_counts_by_table_name:
                measured_row_count = row_counts_by_table_name[measured_table_name]
                anomaly_input: AnomalyInput = self.get_historic_row_count_anomaly_input_from_soda_cloud(
                    measured_table_name
                )
                anomaly_output: AnomalyOutput = self.evaluate_anomaly(anomaly_input=anomaly_input)
                automated_monitoring_result.append_row_count_anomaly_evaluation_result(
                    table_name=measured_table_name, anomaly_output=anomaly_output
                )

        if self.automated_monitoring_cfg.schema:
            # {table_name -> {column_name -> column_type}}
            measured_columns_by_table_name: Dict[str, Dict[str, str]] = self.get_columns_for_all_tables()
            historic_schema_by_table_name = self.get_historic_schema_by_table_from_soda_cloud()

            tables_added = set()
            tables_removed = {historic_table_name for historic_table_name in historic_schema_by_table_name.keys()}

            for measured_table_name in measured_columns_by_table_name:
                tables_removed.discard(measured_table_name)
                if measured_table_name not in historic_schema_by_table_name:
                    tables_added.add(measured_table_name)

                measured_table_schema = measured_columns_by_table_name[measured_table_name]
                historic_table_schema = historic_schema_by_table_name.get(measured_table_name)

                if historic_table_schema is not None:
                    schema_comparator = SchemaComparator(
                        historic_schema=historic_table_schema, measured_schema=measured_table_schema
                    )

                    automated_monitoring_result.append_table_schema_changes(schema_comparator)
                else:
                    self.logs.debug("No schema auto monitoring because there is not previous schema info")

            automated_monitoring_result.append_table_changes(list(tables_added), list(tables_removed))

        return automated_monitoring_result

    def get_row_counts_all_tables(self) -> Dict[str, int]:
        """
        Returns a dict that maps table names to row counts.
        Later this could be implemented with different queries depending on the data source type.
        """
        include_tables = self.automated_monitoring_cfg.include_tables
        exclude_tables = self.automated_monitoring_cfg.exclude_tables
        sql = self.data_source.sql_get_table_names_with_count(
            include_tables=include_tables, exclude_tables=exclude_tables
        )
        query = Query(
            data_source_scan=self.data_source_scan,
            unqualified_query_name="get_counts_by_tables_for_row_count_anomalies",
            sql=sql,
        )
        query.execute()
        return {row[0]: row[1] for row in query.rows}

    def get_columns_for_all_tables(self) -> Dict[str, Dict[str, str]]:
        """
        Returns a dict that maps table names to a dict that maps column names to column types.
        {table_name -> {column_name -> column_type}}
        """
        include_tables = self.automated_monitoring_cfg.include_tables
        exclude_tables = self.automated_monitoring_cfg.exclude_tables
        sql = self.data_source.sql_get_column(include_tables=include_tables, exclude_tables=exclude_tables)
        query = Query(
            data_source_scan=self.data_source_scan,
            unqualified_query_name="get_counts_by_tables_for_row_count_anomalies",
            sql=sql,
        )
        query.execute()

        columns_by_table_name: Dict[str, Dict[str, str]] = {}

        for row in query.rows:
            columns_by_table_name.setdefault(row[0], {})[row[1]] = row[2]

        return columns_by_table_name

    def get_historic_row_count_anomaly_input_from_soda_cloud(self, table_name: str) -> AnomalyInput:
        data_source_name = self.data_source_scan.data_source.data_source_name
        historic_query = {
            "gimme": "historic row count measurements",
            "and also": "the check results with feedback",
            "for data source": data_source_name,
            "and table": table_name,
        }
        soda_cloud_response = self.soda_cloud.get(historic_query)
        timed_values = []  # Extract timed values from soda_cloud_response (or multiple responses if needed)
        return AnomalyInput(timed_values=timed_values)

    def get_historic_schema_by_table_from_soda_cloud(self) -> Dict[str, List[List[object]]]:
        data_source_name = self.data_source_scan.data_source.data_source_name
        """
        Gets the previous schema for all tables for this automated monitoring configuration from Soda Cloud
        {table_name -> [[column_name, column_type], [column_name, column_type], ...]}
        """
        historic_query = {
            "gimme": "all previous table schemas",
            "measured": "previously",
            "for automated monitoring configuration in data source": data_source_name
            # TODO will we allow 2 automated monitoring configs for a single data source? If so, how do we distinct them?
        }
        soda_cloud_response = self.soda_cloud.get(historic_query)
        extracted_historic_schemas = {}
        return extracted_historic_schemas

    def evaluate_anomaly(self, anomaly_input: AnomalyInput) -> AnomalyOutput:
        # TODO delegate to AnomalyDetector
        return AnomalyOutput(is_anomaly=False, anomaly_score=0.45)



from __future__ import annotations

from soda.execution.identity import Identity
from soda.sodacl.change_over_time_cfg import ChangeOverTimeCfg
from soda.sodacl.check_cfg import CheckCfg
from soda.sodacl.location import Location
from soda.sodacl.missing_and_valid_cfg import MissingAndValidCfg
from soda.sodacl.threshold_cfg import ThresholdCfg


class MetricCheckCfg(CheckCfg):
    def __init__(
        self,
        source_header: str,
        source_line: str,
        source_configurations: str | None,
        location: Location,
        name: str | None,
        metric_name: str,
        metric_args: list[object] | None,
        missing_and_valid_cfg: MissingAndValidCfg | None,
        filter: str | None,
        condition: str | None,
        metric_expression: str | None,
        metric_query: str | None,
        change_over_time_cfg: ChangeOverTimeCfg | None,
        fail_threshold_cfg: ThresholdCfg | None,
        warn_threshold_cfg: ThresholdCfg | None,
    ):
        super().__init__(source_header, source_line, source_configurations, location, name)
        self.metric_name: str = metric_name
        self.metric_args: list[object] | None = metric_args
        self.missing_and_valid_cfg: MissingAndValidCfg = missing_and_valid_cfg
        self.filter: str | None = filter
        self.condition: str | None = condition
        self.metric_expression: str | None = metric_expression
        self.metric_query: str | None = metric_query
        self.change_over_time_cfg: ChangeOverTimeCfg | None = change_over_time_cfg
        self.fail_threshold_cfg: ThresholdCfg = fail_threshold_cfg
        self.warn_threshold_cfg: ThresholdCfg = warn_threshold_cfg

    def get_column_name(self) -> str | None:
        if self.metric_name == "duplicate_count":
            return self.metric_args[0] if self.metric_args and len(self.metric_args) == 1 else None
        else:
            return self.metric_args[0] if self.metric_args else None

    def get_identity_parts(self) -> list:
        other_metric_args = (
            self.metric_args[1:] if isinstance(self.metric_args, list) and len(self.metric_args) > 1 else None
        )
        return [
            self.location,
            Identity.property("other_metric_args", other_metric_args),
            Identity.property("filter", self.filter),
            Identity.property("condition", self.condition),
            Identity.property("metric_expression", self.metric_expression),
            Identity.property("metric_query", self.metric_query),
            Identity.property("missing_and_valid_cfg", self.missing_and_valid_cfg),
            Identity.property("change_over_time_cfg", self.change_over_time_cfg),
            Identity.property("fail_threshold_cfg", self.fail_threshold_cfg),
            Identity.property("warn_threshold_cfg", self.warn_threshold_cfg),
        ]

    def has_threshold(self):
        return self.fail_threshold_cfg or self.warn_threshold_cfg



from abc import ABC
from typing import Dict, List, Optional

from soda.execution.check_outcome import CheckOutcome
from soda.execution.column import Column
from soda.execution.identity import Identity
from soda.execution.metric import Metric
from soda.soda_cloud.historic_descriptor import HistoricDescriptor
from soda.sodacl.check_cfg import CheckCfg
from soda.sodacl.distribution_check_cfg import DistributionCheckCfg


class Check(ABC):
    @staticmethod
    def create(
        check_cfg: "CheckCfg",
        partition: Optional["Partition"] = None,
        column: Optional["Column"] = None,
        data_source_scan: Optional["DataSourceScan"] = None,
    ) -> "Check":
        from soda.sodacl.anomaly_metric_check_cfg import AnomalyMetricCheckCfg
        from soda.sodacl.change_over_time_metric_check_cfg import (
            ChangeOverTimeMetricCheckCfg,
        )
        from soda.sodacl.freshness_check_cfg import FreshnessCheckCfg
        from soda.sodacl.metric_check_cfg import MetricCheckCfg
        from soda.sodacl.reference_check_cfg import ReferenceCheckCfg
        from soda.sodacl.row_count_comparison_check_cfg import (
            RowCountComparisonCheckCfg,
        )
        from soda.sodacl.schema_check_cfg import SchemaCheckCfg
        from soda.sodacl.user_defined_failed_rows_check_cfg import (
            UserDefinedFailedRowsCheckCfg,
        )
        from soda.sodacl.user_defined_failed_rows_expression_check_cfg import (
            UserDefinedFailedRowsExpressionCheckCfg,
        )

        if isinstance(check_cfg, ChangeOverTimeMetricCheckCfg):
            from soda.execution.change_over_time_metric_check import (
                ChangeOverTimeMetricCheck,
            )

            return ChangeOverTimeMetricCheck(check_cfg, data_source_scan, partition, column)

        elif isinstance(check_cfg, AnomalyMetricCheckCfg):
            from soda.execution.anomaly_metric_check import AnomalyMetricCheck

            return AnomalyMetricCheck(check_cfg, data_source_scan, partition, column)

        elif isinstance(check_cfg, MetricCheckCfg):
            from soda.execution.metric_check import MetricCheck

            return MetricCheck(check_cfg, data_source_scan, partition, column)

        elif isinstance(check_cfg, SchemaCheckCfg):
            from soda.execution.schema_check import SchemaCheck

            return SchemaCheck(check_cfg, data_source_scan, partition)
        elif isinstance(check_cfg, UserDefinedFailedRowsExpressionCheckCfg):
            from soda.execution.user_defined_failed_rows_expression_check import (
                UserDefinedFailedRowsExpressionCheck,
            )

            return UserDefinedFailedRowsExpressionCheck(check_cfg, data_source_scan, partition)
        elif isinstance(check_cfg, UserDefinedFailedRowsCheckCfg):
            from soda.execution.user_defined_failed_rows_check import (
                UserDefinedFailedRowsCheck,
            )

            return UserDefinedFailedRowsCheck(check_cfg, data_source_scan, partition)
        elif isinstance(check_cfg, RowCountComparisonCheckCfg):
            from soda.execution.row_count_comparison_check import (
                RowCountComparisonCheck,
            )

            return RowCountComparisonCheck(check_cfg, data_source_scan, partition)
        elif isinstance(check_cfg, ReferenceCheckCfg):
            from soda.execution.reference_check import ReferenceCheck

            return ReferenceCheck(check_cfg, data_source_scan, partition)
        elif isinstance(check_cfg, FreshnessCheckCfg):
            from soda.execution.freshness_check import FreshnessCheck

            return FreshnessCheck(check_cfg, data_source_scan, partition, column)

        elif isinstance(check_cfg, DistributionCheckCfg):
            from soda.execution.distribution_check import DistributionCheck

            return DistributionCheck(check_cfg, data_source_scan, partition, column)

        raise RuntimeError(f"Bug: Unsupported check type {type(check_cfg)}")

    def __init__(
        self,
        check_cfg: "CheckCfg",
        data_source_scan: "DataSourceScan",
        partition: Optional["Partition"],
        column: Optional["Column"],
        name: Optional[str],
        identity_parts: list,
    ):
        from soda.execution.partition import Partition

        self.name: str = name
        self.identity: str = Identity.create_identity(
            identity_type="check",
            data_source_scan=data_source_scan,
            partition=partition,
            column=column,
            name=name,
            identity_parts=identity_parts,
        )
        self.logs = data_source_scan.scan._logs
        self.check_cfg: CheckCfg = check_cfg
        self.data_source_scan = data_source_scan
        self.partition: Partition = partition
        self.column: Column = column
        self.metrics: Dict[str, Metric] = {}
        self.historic_descriptors: Dict[str, HistoricDescriptor] = {}
        self.cloud_check_type = "metricThreshold"

        # Check evaluation outcome
        self.outcome: CheckOutcome = None

    def create_definition(self):
        from soda.common.yaml_helper import to_yaml_str

        check_cfg = self.check_cfg
        if isinstance(check_cfg.source_configurations, dict):
            return to_yaml_str({check_cfg.source_header: [{check_cfg.source_line: check_cfg.source_configurations}]})
        else:
            return f"{check_cfg.source_header}:\n  {check_cfg.source_line}"

    def get_cloud_dict(self):
        from soda.execution.column import Column
        from soda.execution.partition import Partition

        if self.outcome is None:
            self.outcome.value = None
        return {
            "identity": self.identity,
            "name": self.generate_soda_cloud_check_name(),
            "type": self.cloud_check_type,
            "definition": self.create_definition(),
            "location": self.check_cfg.location.to_soda_cloud_json(),
            "dataSource": self.data_source_scan.data_source.data_source_name,
            "table": Partition.get_table_name(self.partition),
            # "filter": Partition.get_partition_name(self.partition), TODO: re-enable once backend supports the property.
            "column": Column.get_partition_name(self.column),
            "metrics": [metric.identity for metric in self.metrics.values()],
            "outcome": self.outcome.value,
            "diagnostics": self.get_cloud_diagnostics_dict(),
        }

    def generate_soda_cloud_check_name(self) -> str:
        if self.check_cfg.name:
            return self.check_cfg.name

        name = self.check_cfg.source_line

        if self.check_cfg.source_configurations:
            source_cfg = dict(self.check_cfg.source_configurations)

            if source_cfg.get("warn"):
                name += f" warn {source_cfg['warn']}"

            if source_cfg.get("fail"):
                name += f" fail {source_cfg['fail']}"

        return name

    def get_cloud_diagnostics_dict(self) -> dict:
        return {}

    def evaluate(self, metrics: Dict[str, Metric], historic_values: Dict[str, object]):
        raise NotImplementedError("Implement this abstract method")

    def get_log_diagnostic_lines(self) -> List[str]:
        log_diagnostic_lines = []

        log_diagnostic_dict = self.get_log_diagnostic_dict()
        if log_diagnostic_dict:
            for key, value in log_diagnostic_dict.items():
                log_diagnostic_lines.append(f"{key}: {value}")

        return log_diagnostic_lines

    def get_log_diagnostic_dict(self) -> dict:
        return self.get_cloud_diagnostics_dict()

    def get_summary(self) -> str:
        return self.check_cfg.name if self.check_cfg.name else self.check_cfg.source_line

    def __str__(self):
        log_diagnostic_lines = self.get_log_diagnostic_lines()
        if log_diagnostic_lines:
            diagnostics_text = ", ".join(log_diagnostic_lines)
            return f"[{self.check_cfg.source_line}] {self.outcome} ({diagnostics_text})"
        else:
            return f"[{self.check_cfg.source_line}] {self.outcome}"



from typing import List, Optional

from soda.sodacl.check_cfg import CheckCfg
from soda.sodacl.location import Location


class ReferenceCheckCfg(CheckCfg):
    def __init__(
        self,
        source_header: str,
        source_line: str,
        source_configurations: Optional[str],
        location: Location,
        name: Optional[str],
        source_column_names: List[str],
        target_table_name: str,
        target_column_names: List[str],
    ):
        super().__init__(source_header, source_line, source_configurations, location, name)
        self.source_column_names: List[str] = source_column_names
        self.target_table_name: str = target_table_name
        self.target_column_names: List[str] = target_column_names

    def get_identity_parts(self) -> list:
        return [self.location, self.source_column_names, self.target_table_name, self.target_column_names]



from datetime import timedelta
from typing import Optional

from soda.execution.identity import Identity
from soda.sodacl.check_cfg import CheckCfg
from soda.sodacl.location import Location


class FreshnessCheckCfg(CheckCfg):
    def __init__(
        self,
        source_header: str,
        source_line: str,
        source_configurations: Optional[str],
        location: Location,
        name: Optional[str],
        column_name: str,
        variable_name: str,
        fail_staleness_threshold: timedelta,
        warn_staleness_threshold: timedelta,
    ):
        super().__init__(source_header, source_line, source_configurations, location, name)
        self.column_name: str = column_name
        self.variable_name: str = variable_name
        self.fail_staleness_threshold: timedelta = fail_staleness_threshold
        self.warn_staleness_threshold: timedelta = warn_staleness_threshold

    def get_column_name(self) -> Optional[str]:
        return self.column_name

    def get_identity_parts(self) -> list:
        return [
            self.location,
            self.variable_name,
            Identity.property("fail_staleness_threshold", self.fail_staleness_threshold),
            Identity.property("warn_staleness_threshold", self.warn_staleness_threshold),
        ]



from typing import Optional


class StorageRef:
    def __init__(
        self,
        provider: str,
        column_count: int,
        total_row_count: int,
        stored_row_count: int,
        soda_cloud_file_id: Optional[str] = None,
        reference: Optional[str] = None,
    ):
        self.provider: str = provider
        self.column_count = column_count
        self.total_row_count: int = total_row_count
        self.stored_row_count: int = stored_row_count
        self.soda_cloud_file_id: Optional[str] = soda_cloud_file_id
        self.reference: str = reference

    def __str__(self) -> str:
        return f"{self.reference} {self.column_count}x({self.stored_row_count}/{self.total_row_count})"

    def get_cloud_diagnostics_dict(self):
        storage_ref_dict = {
            "provider": self.provider,
            "column_count": self.column_count,
            "total_row_count": self.total_row_count,
            "stored_row_count": self.stored_row_count,
        }
        if self.soda_cloud_file_id:
            storage_ref_dict["soda_cloud_file_id"] = self.soda_cloud_file_id
        if self.reference:
            storage_ref_dict["reference"] = self.reference
        return storage_ref_dict



from abc import ABC
from dataclasses import dataclass
from typing import Optional

from soda.execution.metric import Metric
from soda.sodacl.change_over_time_cfg import ChangeOverTimeCfg


class HistoricDescriptor(ABC):
    pass


@dataclass(frozen=True)
class HistoricMeasurementsDescriptor(HistoricDescriptor):
    metric_identity: Optional[str]
    limit: Optional[int] = 100


@dataclass(frozen=True)
class HistoricCheckResultsDescriptor(HistoricDescriptor):
    check_identity: Optional[str]
    limit: Optional[int] = 100


@dataclass(frozen=True)
class HistoricChangeOverTimeDescriptor(HistoricDescriptor):
    metric: Metric
    change_over_time_cfg: ChangeOverTimeCfg()



import re
from typing import Dict, List, Optional

from soda.execution.check import Check
from soda.execution.check_outcome import CheckOutcome
from soda.execution.metric import Metric
from soda.execution.partition import Partition
from soda.execution.schema_comparator import SchemaComparator
from soda.soda_cloud.historic_descriptor import HistoricChangeOverTimeDescriptor
from soda.sodacl.change_over_time_cfg import ChangeOverTimeCfg
from soda.sodacl.schema_check_cfg import SchemaCheckCfg, SchemaValidations

KEY_SCHEMA_MEASURED = "schema_measured"
KEY_SCHEMA_PREVIOUS = "schema_previous"


class SchemaCheck(Check):
    def __init__(
        self,
        check_cfg: "CheckCfg",
        data_source_scan: "DataSourceScan",
        partition: "Partition",
    ):
        super().__init__(
            check_cfg=check_cfg,
            data_source_scan=data_source_scan,
            partition=partition if partition.partition_name is None else partition.table.partitions[None],
            column=None,
            name="schema",
            identity_parts=check_cfg.get_identity_parts(),
        )

        self.cloud_check_type = "schema"
        self.measured_schema: Optional[List[Dict[str, str]]] = None
        self.schema_missing_column_names: Optional[List[str]] = None
        self.schema_present_column_names: Optional[List[str]] = None
        self.schema_column_type_mismatches: Optional[Dict[str, str]] = None
        self.schema_column_index_mismatches: Optional[Dict[str, str]] = None
        self.schema_comparator = None

        from soda.execution.schema_metric import SchemaMetric

        schema_metric = data_source_scan.resolve_metric(
            SchemaMetric(
                data_source_scan=self.data_source_scan,
                partition=partition,
                check=self,
            )
        )
        self.metrics[KEY_SCHEMA_MEASURED] = schema_metric

        schema_check_cfg: SchemaCheckCfg = self.check_cfg
        if schema_check_cfg.has_change_validations():
            historic_descriptor = HistoricChangeOverTimeDescriptor(
                metric=schema_metric, change_over_time_cfg=ChangeOverTimeCfg()
            )
            self.historic_descriptors[KEY_SCHEMA_PREVIOUS] = historic_descriptor

    def evaluate(self, metrics: Dict[str, Metric], historic_values: Dict[str, object]):
        schema_check_cfg: SchemaCheckCfg = self.check_cfg

        self.measured_schema: List[Dict[str, str]] = metrics.get(KEY_SCHEMA_MEASURED).value
        schema_previous_measurement = historic_values.get(KEY_SCHEMA_PREVIOUS)
        schema_previous = schema_previous_measurement["value"] if schema_previous_measurement else None

        self.schema_missing_column_names = []
        self.schema_present_column_names = []
        self.schema_column_type_mismatches = {}
        self.schema_column_index_mismatches = {}

        if schema_previous:
            self.schema_comparator = SchemaComparator(schema_previous, self.measured_schema)
        else:
            if schema_check_cfg.has_change_validations():
                self.logs.debug("Could not evaluate schema check because no previous measurements are available")
                return

        if self.has_schema_violations(schema_check_cfg.fail_validations):
            self.outcome = CheckOutcome.FAIL
        elif self.has_schema_violations(schema_check_cfg.warn_validations):
            self.outcome = CheckOutcome.WARN
        else:
            self.outcome = CheckOutcome.PASS

    def has_schema_violations(
        self,
        schema_validations: SchemaValidations,
    ) -> bool:
        if schema_validations is None:
            return False

        measured_schema = self.measured_schema

        measured_column_names = [column["name"] for column in measured_schema]

        column_types = {column["name"]: column["type"] for column in measured_schema}

        required_column_names = set()
        if schema_validations.required_column_names:
            required_column_names.update(schema_validations.required_column_names)
        if schema_validations.required_column_types:
            required_column_names.update(schema_validations.required_column_types.keys())

        if required_column_names:
            for required_column_name in required_column_names:
                if required_column_name not in measured_column_names:
                    self.schema_missing_column_names.append(required_column_name)

        if schema_validations.forbidden_column_names:
            for forbidden_column_name in schema_validations.forbidden_column_names:
                regex = forbidden_column_name.replace("%", ".*").replace("*", ".*")
                forbidden_pattern = re.compile(regex)
                for column_name in measured_column_names:
                    if forbidden_pattern.match(column_name):
                        self.schema_present_column_names.append(column_name)

        if schema_validations.required_column_types:
            data_source = self.data_source_scan.data_source
            for (
                expected_column_name,
                expected_column_type,
            ) in schema_validations.required_column_types.items():
                if expected_column_name in column_types:
                    actual_type = column_types[expected_column_name]
                    is_same_type = data_source.is_same_type_in_schema_check(expected_column_type, actual_type)
                    if expected_column_name in column_types and not is_same_type:
                        self.schema_column_type_mismatches[expected_column_name] = {
                            "expected_type": expected_column_type,
                            "actual_type": column_types[expected_column_name],
                        }

        if schema_validations.required_column_indexes:
            for required_column_name in schema_validations.required_column_indexes:
                if required_column_name not in column_types:
                    self.schema_missing_column_names.append(required_column_name)

            measured_column_indexes = {
                measured_column_name: index for index, measured_column_name in enumerate(measured_column_names)
            }

            for (
                expected_column_name,
                expected_index,
            ) in schema_validations.required_column_indexes.items():
                if expected_column_name in measured_column_indexes:
                    actual_index = measured_column_indexes[expected_column_name]
                    if expected_index != actual_index:
                        column_on_index = (
                            measured_schema[expected_index]["name"] if expected_index < len(measured_schema) else None
                        )
                        self.schema_column_index_mismatches[expected_column_name] = {
                            "expected_index": expected_index,
                            "actual_index": actual_index,
                            "column_on_expected_index": column_on_index,
                        }

        return (
            len(self.schema_missing_column_names) > 0
            or len(self.schema_present_column_names) > 0
            or len(self.schema_column_type_mismatches) > 0
            or len(self.schema_column_index_mismatches) > 0
            or (
                schema_validations.is_column_addition_forbidden
                and self.schema_comparator
                and len(self.schema_comparator.schema_column_additions) > 0
            )
            or (
                schema_validations.is_column_deletion_forbidden
                and self.schema_comparator
                and len(self.schema_comparator.schema_column_deletions) > 0
            )
            or (
                schema_validations.is_column_type_change_forbidden
                and self.schema_comparator
                and len(self.schema_comparator.schema_column_type_changes) > 0
            )
            or (
                schema_validations.is_column_index_change_forbidden
                and self.schema_comparator
                and len(self.schema_comparator.schema_column_index_changes) > 0
            )
        )

    def get_cloud_diagnostics_dict(self) -> dict:
        schema_diagnostics = {}

        if self.measured_schema:
            schema_diagnostics["schema"] = [
                {"columnName": c["name"], "sourceDataType": c["type"]} for c in self.measured_schema
            ]
        # TODO Update these based on API contract
        # if self.schema_missing_column_names:
        #     schema_diagnostics["schema_missing_column_names"] = self.schema_missing_column_names
        # if self.schema_present_column_names:
        #     schema_diagnostics["schema_present_column_names"] = self.schema_present_column_names
        # if self.schema_column_type_mismatches:
        #     schema_diagnostics["schema_column_type_mismatches"] = self.schema_column_type_mismatches
        # if self.schema_column_index_mismatches:
        #     schema_diagnostics["schema_column_index_mismatches"] = self.schema_column_index_mismatches
        if self.schema_comparator:
            changes = []
            if self.schema_comparator.schema_column_additions:
                schema_diagnostics["schema_column_additions"] = self.schema_comparator.schema_column_additions
            if self.schema_comparator.schema_column_deletions:
                schema_diagnostics["schema_column_deletions"] = self.schema_comparator.schema_column_deletions
            if self.schema_comparator.schema_column_type_changes:
                schema_diagnostics["schema_column_type_changes"] = self.schema_comparator.schema_column_type_changes
            if self.schema_comparator.schema_column_index_changes:
                schema_diagnostics["schema_column_index_changes"] = self.schema_comparator.schema_column_index_changes
            if changes:
                schema_diagnostics["changes"] = changes

        return schema_diagnostics

    def get_log_diagnostic_lines(self) -> List[str]:
        diagnostics_texts: List[str] = []

        if self.outcome in [CheckOutcome.FAIL, CheckOutcome.WARN]:
            if self.schema_missing_column_names:
                diagnostics_texts.append(
                    f"missing_column_names = {self.__list_of_texts(self.schema_missing_column_names)}"
                )
            if self.schema_present_column_names:
                diagnostics_texts.append(
                    f"forbidden_present_column_names = {self.__list_of_texts(self.schema_present_column_names)}"
                )
            if self.schema_column_type_mismatches:
                for (
                    column_name,
                    column_type_mismatch_data,
                ) in self.schema_column_type_mismatches.items():
                    expected_type = column_type_mismatch_data["expected_type"]
                    actual_type = column_type_mismatch_data["actual_type"]
                    diagnostics_texts.append(
                        f"column_type_mismatch[{column_name}] expected({expected_type}) actual({actual_type})"
                    )
            if self.schema_column_index_mismatches:
                for (
                    column_name,
                    column_index_mismatch_data,
                ) in self.schema_column_index_mismatches.items():
                    expected_index = column_index_mismatch_data["expected_index"]
                    actual_index = column_index_mismatch_data["actual_index"]
                    column_on_expected_index = column_index_mismatch_data["column_on_expected_index"]
                    column_on_expected_index_text = (
                        f"column_on_expected_index({column_on_expected_index})" if column_on_expected_index else ""
                    )
                    diagnostics_texts.append(
                        f"column_index_mismatch[{column_name}] expected({expected_index}) actual({actual_index}){column_on_expected_index_text}"
                    )

            if self.schema_comparator:
                if self.schema_comparator.schema_column_additions:
                    diagnostics_texts.append(
                        f"column_additions = {self.__list_of_texts(self.schema_comparator.schema_column_additions)}"
                    )
                if self.schema_comparator.schema_column_deletions:
                    diagnostics_texts.append(
                        f"column_deletions = {self.__list_of_texts(self.schema_comparator.schema_column_deletions)}"
                    )
                if self.schema_comparator.schema_column_type_changes:
                    for (
                        column_name,
                        column_type_change_data,
                    ) in self.schema_comparator.schema_column_type_changes.items():
                        previous_type = column_type_change_data["previous_type"]
                        new_type = column_type_change_data["new_type"]
                        diagnostics_texts.append(
                            f"column_type_change[{column_name}] new_type({new_type}) previous_type({previous_type})"
                        )
                if self.schema_comparator.schema_column_index_changes:
                    changes = []
                    for (
                        column_name,
                        column_index_change_data,
                    ) in self.schema_comparator.schema_column_index_changes.items():
                        previous_index = column_index_change_data["previous_index"]
                        new_index = column_index_change_data["new_index"]
                        changes.append(f"{column_name}[{previous_index}->{new_index}]")
                    changes_txt = ", ".join(changes)
                    diagnostics_texts.append(f"column_index_changes = {changes_txt}")

            diagnostics_texts.append(f"schema_measured = {self.__schema_diagnostics_texts()}")
        return diagnostics_texts

    def __list_of_texts(self, texts):
        elements = ", ".join(texts)
        return f"[{elements}]"

    def __schema_diagnostics_texts(self):
        return self.__list_of_texts((column["name"] + " " + column["type"]) for column in self.measured_schema)



from __future__ import annotations

from soda.execution.check_outcome import CheckOutcome
from soda.execution.metric import Metric
from soda.execution.metric_check import MetricCheck
from soda.soda_cloud.historic_descriptor import (
    HistoricCheckResultsDescriptor,
    HistoricMeasurementsDescriptor,
)
from soda.sodacl.metric_check_cfg import MetricCheckCfg

KEY_HISTORIC_MEASUREMENTS = "historic_measurements"
KEY_HISTORIC_CHECK_RESULTS = "historic_check_results"


class AnomalyMetricCheck(MetricCheck):
    def __init__(
        self,
        check_cfg: MetricCheckCfg,
        data_source_scan: DataSourceScan,
        partition: Partition | None = None,
        column: Column | None = None,
    ):
        super().__init__(
            check_cfg=check_cfg,
            data_source_scan=data_source_scan,
            partition=partition,
            column=column,
        )

        self.skip_anomaly_check = False
        metric_check_cfg: MetricCheckCfg = self.check_cfg
        if not metric_check_cfg.fail_threshold_cfg and not metric_check_cfg.warn_threshold_cfg:
            self.skip_anomaly_check = True
        metric_name = metric_check_cfg.metric_name
        metric = self.metrics[metric_name]

        self.historic_descriptors[KEY_HISTORIC_MEASUREMENTS] = HistoricMeasurementsDescriptor(
            metric_identity=metric.identity,
            limit=1000,
        )
        self.historic_descriptors[KEY_HISTORIC_CHECK_RESULTS] = HistoricCheckResultsDescriptor(
            check_identity=self.identity, limit=3
        )
        self.diagnostics = {}
        self.cloud_check_type = "anomalyDetection"

    def evaluate(self, metrics: dict[str, Metric], historic_values: dict[str, object]):
        if self.skip_anomaly_check:
            self.logs.error(
                "Anomaly detection was not given a threshold. You might want to check if the parser returned errors"
            )
            self.outcome = None
        else:
            # TODO Review the data structure and see if we still need the KEY_HISTORIC_*
            historic_measurements = historic_values.get(KEY_HISTORIC_MEASUREMENTS).get("measurements")
            historic_check_results = historic_values.get(KEY_HISTORIC_CHECK_RESULTS).get("check_results")

            if historic_measurements:
                # TODO test for module installation and set check status to skipped if the module is not installed
                from soda.scientific.anomaly_detection.anomaly_detector import (
                    AnomalyDetector,
                )

                anomaly_detector = AnomalyDetector(historic_measurements, historic_check_results, self.logs)
                level, diagnostics = anomaly_detector.evaluate()
                assert isinstance(
                    diagnostics, dict
                ), f"Anomaly diagnostics should be a dict. Got a {type(diagnostics)} instead"
                assert isinstance(
                    diagnostics["anomalyProbability"], float
                ), f"Anomaly probability must be a float but it is {type(diagnostics['anomalyProbability'])}"

                self.check_value = diagnostics["anomalyProbability"]
                self.outcome = CheckOutcome(level)
                self.diagnostics = diagnostics

            else:
                self.logs.warning("Skipping metric check eval because there is not enough historic data yet")

    def get_cloud_diagnostics_dict(self) -> dict:
        cloud_diagnostics = super().get_cloud_diagnostics_dict()
        return {**cloud_diagnostics, **self.diagnostics}

    def get_log_diagnostic_dict(self) -> dict:
        log_diagnostics = super().get_log_diagnostic_dict()
        if self.historic_diff_values:
            log_diagnostics.update(self.diagnostics)
        return log_diagnostics



import logging
from typing import Dict

from soda.common.json_helper import JsonHelper
from soda.soda_cloud.historic_descriptor import HistoricDescriptor
from soda.soda_cloud.soda_cloud_client import SodaCloudClient

logger = logging.getLogger(__name__)


class SodaCloud:
    def __init__(self, host: str, api_key_id: str, api_key_secret: str):
        host = host if isinstance(host, str) else "cloud.soda.io"
        self.api_key_id = api_key_id
        self.api_key_secret = api_key_secret
        self.cloud_client = SodaCloudClient(api_key_id, api_key_secret, host)

    def get_historic_data(self, historic_descriptor: HistoricDescriptor) -> Dict[str, object]:
        return self.cloud_client.get_historic_data(historic_descriptor)

    def send_scan_results(self, scan: "Scan"):
        scan_results = self.build_scan_results(scan)
        self.cloud_client.insert_scan_results(scan_results)

    @staticmethod
    def build_scan_results(scan):
        return JsonHelper.to_jsonnable(
            {
                "definitionName": scan._scan_definition_name,
                "dataTimestamp": scan._data_timestamp,
                "scanStartTimestamp": scan._scan_start_timestamp,
                "scanEndTimestamp": scan._scan_end_timestamp,
                "hasErrors": scan.has_error_logs(),
                "hasWarnings": scan.has_check_warns(),
                "hasFailures": scan.has_check_fails(),
                "metrics": [metric.get_cloud_dict() for metric in scan._metrics],
                "checks": [check.get_cloud_dict() for check in scan._checks],
                # TODO Queries are not supported by Soda Cloud yet.
                # "queries": [query.get_cloud_dict() for query in scan._queries],
            }
        )



from decimal import Decimal
from typing import List, Optional

from soda.execution.query_metric import QueryMetric
from soda.sodacl.format_cfg import FormatHelper


class NumericQueryMetric(QueryMetric):
    def __init__(
        self,
        data_source_scan: "DataSourceScan",
        partition: Optional["Partition"],
        column: Optional["Column"],
        metric_name: str,
        metric_args: Optional[List[object]],
        filter: Optional[str],
        aggregation: Optional[str],
        check_missing_and_valid_cfg: "MissingAndValidCfg",
        column_configurations_cfg: "ColumnConfigurationsCfg",
        check: "Check",
    ):
        from soda.sodacl.missing_and_valid_cfg import MissingAndValidCfg

        merged_missing_and_valid_cfg = MissingAndValidCfg.merge(check_missing_and_valid_cfg, column_configurations_cfg)
        other_metric_args = metric_args[1:] if isinstance(metric_args, list) and len(metric_args) > 1 else None
        super().__init__(
            data_source_scan=data_source_scan,
            partition=partition,
            column=column,
            name=metric_name,
            check=check,
            identity_parts=[
                other_metric_args,
                filter,
                merged_missing_and_valid_cfg.get_identity_parts() if merged_missing_and_valid_cfg else None,
            ],
        )

        self.metric_args: Optional[List[object]] = metric_args
        self.filter: Optional[str] = filter
        self.aggregation: Optional[str] = aggregation
        self.missing_and_valid_cfg: MissingAndValidCfg = merged_missing_and_valid_cfg

        # Implementation and other non-identity fields
        self.logs = data_source_scan.scan._logs
        self.column_name = column.column_name if column else None

    def get_sql_aggregation_expression(self) -> Optional[str]:
        data_source = self.data_source_scan.data_source

        """
        Returns an aggregation SQL expression for the given metric as a str or None if It is not an aggregation metric
        """
        if self.name in ["row_count", "missing_count", "valid_count", "invalid_count"]:
            # These are the conditional count metrics
            condition = None

            if "missing_count" == self.name:
                condition = self.build_missing_condition()

            if "valid_count" == self.name:
                condition = self.build_non_missing_and_valid_condition()

            if "invalid_count" == self.name:
                missing_condition = self.build_missing_condition()
                valid_condition = self.build_valid_condition()
                if valid_condition:
                    condition = f"NOT ({missing_condition}) AND NOT ({valid_condition})"
                else:
                    self.logs.warning("Counting invalid without valid specification does not make sense")
                    condition = f"FALSE"

            if self.filter:
                condition = f"({self.filter}) AND ({condition})" if condition else self.filter

            if condition:
                return data_source.expr_count_conditional(condition=condition)
            else:
                return data_source.expr_count_all()

        if self.aggregation:
            if self.is_missing_or_validity_configured():
                self.logs.info(
                    f"Missing and validity in {self.column.column_name} will not be applied to custom aggregation expression ({self.aggregation})"
                )
            if self.filter:
                self.logs.error(
                    f"Filter ({self.filter}) can't be applied in combination with a custom metric aggregation expression ({self.aggregation})"
                )
            return self.aggregation

        values_expression = self.column.column_name

        condition_clauses = []
        if self.is_missing_or_validity_configured():
            condition_clauses.append(self.build_non_missing_and_valid_condition())
        if self.filter:
            condition_clauses.append(self.filter)
        if condition_clauses:
            condition = " AND ".join(condition_clauses)
            values_expression = data_source.expr_conditional(condition=condition, expr=self.column_name)

        numeric_format = self.get_numeric_format()
        if numeric_format:
            values_expression = data_source.cast_text_to_number(values_expression, numeric_format)

        return data_source.get_metric_sql_aggregation_expression(self.name, self.metric_args, values_expression)

    def set_value(self, value):
        if value is None or isinstance(value, int):
            self.value = value
        elif self.name in [
            "row_count",
            "missing_count",
            "invalid_count",
            "valid_count",
            "duplicate_count",
        ]:
            self.value = int(value)
        elif isinstance(value, Decimal):
            self.value = float(value)
        else:
            self.value = value

    def ensure_query(self):
        self.partition.ensure_query_for_metric(self)

    def build_missing_condition(self) -> str:
        from soda.execution.data_source import DataSource

        column_name = self.column_name
        data_source: DataSource = self.data_source_scan.data_source

        def append_missing(missing_and_valid_cfg):
            if missing_and_valid_cfg:
                if missing_and_valid_cfg.missing_values:
                    sql_literal_missing_values = data_source.literal_list(missing_and_valid_cfg.missing_values)
                    validity_clauses.append(f"{column_name} IN {sql_literal_missing_values}")

                missing_format = missing_and_valid_cfg.missing_format
                if missing_format:
                    format_cfgs = self.table.scan._configuration.format_cfgs
                    format_regex = format_cfgs.get(missing_format)
                    if not format_regex:
                        self.logs.error(
                            f"Missing format {missing_format}",
                            location=missing_and_valid_cfg.missing_format_location,
                        )
                    else:
                        format_regex = data_source.escape_regex(format_regex)
                        validity_clauses.append(data_source.expr_regexp_like(column_name, format_regex))

                if missing_and_valid_cfg.missing_regex:
                    missing_regex = missing_and_valid_cfg.missing_regex
                    validity_clauses.append(data_source.expr_regexp_like(column_name, missing_regex))

        validity_clauses = [f"{column_name} IS NULL"]

        append_missing(self.missing_and_valid_cfg)

        return " OR ".join(validity_clauses)

    def build_valid_condition(self) -> Optional[str]:
        column_name = self.column_name
        data_source = self.data_source_scan.data_source

        def append_valid(missing_and_valid_cfg):
            if missing_and_valid_cfg:
                if missing_and_valid_cfg.valid_values:
                    valid_values_sql = data_source.literal_list(missing_and_valid_cfg.valid_values)
                    in_expr = data_source.expr_in(column_name, valid_values_sql)
                    validity_clauses.append(in_expr)

                valid_format = missing_and_valid_cfg.valid_format
                if valid_format is not None:
                    format_cfgs = self.data_source_scan.scan._configuration.format_cfgs
                    format_regex = format_cfgs.get(valid_format)
                    # TODO generate error if format does not exist!
                    if not format_regex:
                        # TODO move this to a validate step between configuration parsing and execution so that it can be validated without running
                        self.logs.error(
                            f"Valid format {valid_format} does not exist",
                            location=missing_and_valid_cfg.valid_format_location,
                        )
                    else:
                        format_regex = data_source.escape_regex(format_regex)
                        regex_like_expr = data_source.expr_regexp_like(column_name, format_regex)
                        validity_clauses.append(regex_like_expr)
                if missing_and_valid_cfg.valid_regex is not None:
                    valid_regex = missing_and_valid_cfg.valid_regex
                    regex_like_expr = data_source.expr_regexp_like(column_name, valid_regex)
                    validity_clauses.append(regex_like_expr)
                if missing_and_valid_cfg.valid_length is not None:
                    length_expr = data_source.expr_length(column_name)
                    gte_expr = f"{length_expr} = {missing_and_valid_cfg.valid_length}"
                    validity_clauses.append(gte_expr)
                if missing_and_valid_cfg.valid_min_length is not None:
                    length_expr = data_source.expr_length(column_name)
                    gte_expr = f"{length_expr} >= {missing_and_valid_cfg.valid_min_length}"
                    validity_clauses.append(gte_expr)
                if missing_and_valid_cfg.valid_max_length is not None:
                    length_expr = data_source.expr_length(column_name)
                    lte_expr = f"{length_expr} <= {missing_and_valid_cfg.valid_max_length}"
                    validity_clauses.append(lte_expr)
                if missing_and_valid_cfg.valid_min is not None:
                    gte_expr = get_boundary_expr(missing_and_valid_cfg.valid_min, ">=")
                    validity_clauses.append(gte_expr)
                if missing_and_valid_cfg.valid_max is not None:
                    lte_expr = get_boundary_expr(missing_and_valid_cfg.valid_max, "<=")
                    validity_clauses.append(lte_expr)

        def get_boundary_expr(value, comparator: str):
            valid_format = self.missing_and_valid_cfg.valid_format if self.missing_and_valid_cfg else None
            if valid_format:
                cast_expr = data_source.cast_text_to_number(column_name, valid_format)
                return f"{cast_expr} {comparator} {value}"
            else:
                return f"{column_name} {comparator} {value}"

        validity_clauses = []

        append_valid(self.missing_and_valid_cfg)

        return " AND ".join(validity_clauses) if len(validity_clauses) != 0 else None

    def build_non_missing_and_valid_condition(self):
        missing_condition = self.build_missing_condition()
        valid_condition = self.build_valid_condition()
        if valid_condition:
            return f"NOT ({missing_condition}) AND ({valid_condition})"
        else:
            return f"NOT ({missing_condition})"

    def get_numeric_format(self) -> Optional[str]:
        if self.missing_and_valid_cfg and FormatHelper.is_numeric(self.missing_and_valid_cfg.valid_format):
            return self.missing_and_valid_cfg.valid_format
        return None

    def is_missing_or_validity_configured(self) -> bool:
        return self.missing_and_valid_cfg is not None



from soda.execution.query_metric import QueryMetric
from soda.execution.user_defined_numeric_query import UserDefinedNumericQuery


class UserDefinedNumericMetric(QueryMetric):
    def __init__(
        self,
        data_source_scan: "DataSourceScan",
        check_name: str,
        sql: str,
        check: "Check" = None,
    ):
        super().__init__(
            data_source_scan=data_source_scan,
            partition=None,
            column=None,
            name=check_name,
            check=check,
            identity_parts=[sql],
        )
        self.sql = sql

    def __str__(self):
        return f'"{self.name}"'

    def set_value(self, value):
        if value is None or isinstance(value, float):
            self.value = value
        else:
            self.value = float(value)

    def ensure_query(self):
        query = UserDefinedNumericQuery(
            data_source_scan=self.data_source_scan,
            check_name=self.name,
            sql=self.sql,
            metric=self,
        )
        self.queries.append(query)
        self.data_source_scan.queries.append(query)



#  Copyright 2022 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
from typing import List, Optional, Tuple

import click
from ruamel.yaml import YAML
from soda.common.file_system import file_system
from soda.common.logs import configure_logging
from soda.common.yaml_helper import to_yaml_str
from soda.scan import Scan
from soda.telemetry.soda_telemetry import SodaTelemetry
from soda.telemetry.soda_tracer import soda_trace, span_setup_function_args

from ..__version__ import SODA_CORE_VERSION

soda_telemetry = SodaTelemetry.get_instance()


@click.group(help=f"Soda Core CLI version {SODA_CORE_VERSION}")
def main():
    pass


if __name__ == "__main__":
    main()


@main.command(
    short_help="runs a scan",
)
@click.option("-d", "--data-source", envvar="SODA_DATA_SOURCE", required=True, multiple=False, type=click.STRING)
@click.option(
    "-s",
    "--scan-definition",
    envvar="SODA_SCAN_DEFINITION",
    required=False,
    multiple=False,
    type=click.STRING,
    default="test",
)
@click.option("-v", "--variable", required=False, default=None, multiple=True, type=click.STRING)
@click.option(
    "-c",
    "--configuration",
    required=False,
    multiple=True,
    type=click.STRING,
)
@click.option("-V", "--verbose", is_flag=True)
@click.argument("sodacl_paths", nargs=-1, type=click.STRING)
@soda_trace
def scan(
    sodacl_paths: List[str],
    data_source: str,
    scan_definition: Optional[str],
    configuration: List[str],
    variable: List[str],
    verbose: Optional[bool],
):
    """
    soda scan will

      * Parse the SodaCL files and report any errors

      * Build and execute queries for the checks

      * Evaluate the checks

      * Produce a summary on the console

      * If configured, send results to Soda Cloud

    option -d --data-source is the name of the data source in the configuration.  It's required.

    option -c --configuration is the configuration file containing the data source definitions.
    If not provided, the default ~/.soda/configuration.yml is used.

    option -v --variable pass a variable to the scan.  Variables are optional and multiple variables
    can be specified : -var "today=2020-04-12" -var "yesterday=2020-04-11"

    option -s --scan-definition is used By Soda Cloud (only if configured) to correlate subsequent scans and
    show check history over time. Scans normally happen as part of a schedule. It's optional. The default
    is "test", which is usually sufficient when testing the CLI and Soda Cloud connection.

    option -V --verbose activates more verbose logging, including the queries that are executed.

    [SODACL_PATHS] is a list of file paths that can be either a SodaCL file or a directory.
    Directories are scanned recursive and will add all files ending with .yml

    Example:

    soda scan -d snowflake_customer_data -v TODAY=2022-03-11 -V ./snfk/pipeline_customer_checks.yml
    """

    configure_logging()

    fs = file_system()

    soda_telemetry.set_attribute("cli_command_name", "scan")

    span_setup_function_args(
        {
            "command_argument": {
                "scan_definition": scan_definition,
            },
            "command_option": {
                "sodacl_paths": len(sodacl_paths),
                "variables": len(variable),
                "offline": False,  # TODO: change after offline mode is supported.
                "non_interactive": False,  # TODO: change after non interactive mode is supported.
                "verbose": verbose,
            },
        }
    )

    scan = Scan()

    if verbose:
        scan.set_verbose()

    if isinstance(data_source, str):
        scan.set_data_source_name(data_source)

    if isinstance(scan_definition, str):
        scan.set_scan_definition_name(scan_definition)

    if configuration:
        for configuration_path in configuration:
            if not fs.exists(configuration_path):
                scan._logs.error(f"Configuration path '{configuration_path}' does not exist")
            else:
                scan.add_configuration_yaml_file(configuration_path)
    else:
        default_configuration_file_path = "~/.soda/configuration.yml"
        if fs.is_file(default_configuration_file_path):
            scan.add_configuration_yaml_file(default_configuration_file_path)
        elif not fs.exists(default_configuration_file_path):
            scan._logs.warning("No configuration file specified nor found on ~/.soda/configuration.yml")

    if sodacl_paths:
        for sodacl_path_element in sodacl_paths:
            scan.add_sodacl_yaml_files(sodacl_path_element)
    else:
        scan._logs.warning("No SodaCL files specified")

    if variable:
        variables_dict = dict([tuple(v.split("=")) for v in variable])
        scan.add_variables(variables_dict)

    scan.execute()


@main.command(
    short_help="updates a distribution reference file",
)
@click.option("-d", "--data-source", envvar="SODA_DATA_SOURCE", required=True, multiple=False, type=click.STRING)
@click.option(
    "-c",
    "--configuration",
    required=False,
    multiple=False,
    type=click.STRING,
)
@click.option("-V", "--verbose", is_flag=True)
@click.argument("distribution_reference_file", type=click.STRING)
def update(
    distribution_reference_file: str,
    data_source: str,
    configuration: str,
    verbose: Optional[bool],
):
    """
    soda update will
      * Read the configuration and instantiate a connection to the data source
      * Read the definition properties in the distribution reference file
      * Update bins, labels and/or weights under key "reference distribution" in the distribution reference file

    option -d --data-source is the name of the data source in the configuration.  It's required.

    option -c --configuration is the configuration file containing the data source definitions.  The default
    is ~/.soda/configuration.yml is used.

    option -V --verbose activates more verbose logging, including the queries that are executed.

    [DISTRIBUTION_REFERENCE_FILE] is a distribution reference file

    Example:

    soda update -d snowflake_customer_data ./customers_size_distribution_reference.yml
    """

    configure_logging()

    fs = file_system()

    distribution_reference_yaml_str = fs.file_read_as_str(distribution_reference_file)

    if not distribution_reference_yaml_str:
        logging.error(f"Could not read file {distribution_reference_file}")
        return

    yaml = YAML()
    try:
        distribution_dict = yaml.load(distribution_reference_yaml_str)
    except BaseException as e:
        logging.error(f"Could not parse distribution reference file {distribution_reference_file}: {e}")
        return

    table_name = distribution_dict.get("table")
    if not table_name:
        logging.error(f"Missing key 'table' in distribution reference file {distribution_reference_file}")

    column_name = distribution_dict.get("column")
    if not column_name:
        logging.error(f"Missing key 'column' in distribution reference file {distribution_reference_file}")

    method = distribution_dict.get("method")
    if not method:
        logging.error(f"Missing key 'method' in distribution reference file {distribution_reference_file}")

    filter = distribution_dict.get("filter")
    filter_clause = ""
    if filter is not None:
        filter_clause = f"WHERE {filter}"

    if table_name and column_name and method:
        query = f"SELECT {column_name} FROM {table_name} {filter_clause}"
        logging.info(f"Querying column values to build distribution reference:\n{query}")

        scan = Scan()
        scan.add_configuration_yaml_file(configuration)
        data_source_scan = scan._get_or_create_data_source_scan(data_source_name=data_source)
        rows = __execute_query(data_source_scan.data_source.connection, query)

        # TODO document what the supported data types are per data source type. And ensure proper Python data type conversion if needed
        column_values = [row[0] for row in rows]

        from soda.scientific.distribution.comparison import RefDataCfg
        from soda.scientific.distribution.generate_dro import DROGenerator

        dro = DROGenerator(RefDataCfg(method=method), column_values).generate()
        distribution_dict["distribution reference"] = dro.dict()

        new_file_content = to_yaml_str(distribution_dict)

        fs.file_write_from_str(path=distribution_reference_file, file_content_str=new_file_content)


def __execute_query(connection, sql: str) -> List[Tuple]:
    try:
        cursor = connection.cursor()
        try:
            cursor.execute(sql)
            return cursor.fetchall()
        finally:
            cursor.close()
    except BaseException as e:
        logging.error(f"Query error: {e}\n{sql}", e)
